from .dist import rgb_euclidean, hsl_euclidean, hsv_euclidean, cie76, cie94, ciede2000

from .conv import ( 
    hsl2hsv, hsl2lab, hsl2rgb,
    hsv2hsl, hsv2lab, hsv2rgb,
    rgb2hsl, rgb2hsv, rgb2lab,
)


"""
    sort_palette_by_chan(pal, chan)

    sort_palette_by_chan0(pal): Returns a new palette sorted channel 0. 
    These are aliases for sort_palette_by_chan0:
        sort_palette_by_r
        sort_rgb_palette_by_r
        sort_palette_by_h
        sort_hsv_palette_by_hue
        sort_hsl_palette_by_hue

    sort_palette_by_chan1(pal): Returns a new palette sorted channel 1. 
    These are aliases for sort_palette_by_chan1:
        sort_palette_by_g
        sort_rgb_palette_by_g
        sort_palette_by_s
        sort_hsv_palette_by_sat
        sort_hsl_palette_by_sat

    sort_palette_by_chan2(pal): Returns a new palette sorted channel 2. 
    These are aliases for sort_palette_by_chan2:
        sort_palette_by_b
        sort_rgb_palette_by_b
        sort_palette_by_l
        sort_palette_by_v
        sort_hsv_palette_by_val
        sort_hsl_palette_by_lum

    sort_rgb_palette_by_hue(pal):
    sort_rgb_palette_by_sat(pal):
    sort_rgb_palette_by_val(pal):
    sort_rgb_palette_by_lum(pal):


    sort_hsv_palette_by_lum(pal):
    sort_hsv_palette_by_r(pal):
    sort_hsv_palette_by_g(pal):
    sort_hsv_palette_by_b(pal):    


    sort_hsl_palette_by_val(pal):
    sort_hsl_palette_by_r(pal):
    sort_hsl_palette_by_g(pal):
    sort_hsl_palette_by_b(pal):
    sort_rgb_palette_by_cie76_from(pal, rgb):
    sort_rgb_palette_by_cie94_from(pal, rgb):
    sort_rgb_palette_by_ciede2000_from(pal, rgb):
    sort_rgb_palette_by_rgb_euclidean_from(pal, rgb):
    sort_rgb_palette_by_hsv_euclidean_from(pal, rgb):
    sort_rgb_palette_by_hsl_euclidean_from(pal, rgb):
    sort_hsv_palette_by_cie76_from(pal, hsv):
    sort_hsv_palette_by_cie94_from(pal, hsv):
    sort_hsv_palette_by_ciede2000_from(pal, hsv):
    sort_hsv_palette_by_hsv_euclidean_from(pal, hsv):
    sort_hsl_palette_by_cie76_from(pal, hsl):
    sort_hsl_palette_by_cie94_from(pal, hsl):
    sort_hsl_palette_by_ciede2000_from(pal, hsl):
    sort_hsl_palette_by_hsl_euclidean_from(pal, hsl):
"""

__all__ = [
    'sort_palette_by_chan',
    'sort_palette_by_chan0',
    'sort_palette_by_r',
    'sort_palette_by_h',
    'sort_palette_by_chan1',
    'sort_palette_by_g',
    'sort_palette_by_s',
    'sort_palette_by_chan2',
    'sort_palette_by_b',
    'sort_palette_by_l',
    'sort_palette_by_v',
    'sort_rgb_palette_by_r',
    'sort_rgb_palette_by_g',
    'sort_rgb_palette_by_b',
    'sort_rgb_palette_by_hue',
    'sort_rgb_palette_by_sat',
    'sort_rgb_palette_by_val',
    'sort_rgb_palette_by_lum',
    'sort_hsv_palette_by_hue',
    'sort_hsv_palette_by_sat',
    'sort_hsv_palette_by_val',
    'sort_hsv_palette_by_lum',
    'sort_hsv_palette_by_r',
    'sort_hsv_palette_by_g',
    'sort_hsv_palette_by_b',
    'sort_hsl_palette_by_hue',
    'sort_hsl_palette_by_sat',
    'sort_hsl_palette_by_lum',
    'sort_hsl_palette_by_val',
    'sort_hsl_palette_by_r',
    'sort_hsl_palette_by_g',
    'sort_hsl_palette_by_b',
    'sort_rgb_palette_by_cie76_from',
    'sort_rgb_palette_by_cie94_from',
    'sort_rgb_palette_by_ciede2000_from',
    'sort_rgb_palette_by_rgb_euclidean_from',
    'sort_rgb_palette_by_hsv_euclidean_from',
    'sort_rgb_palette_by_hsl_euclidean_from',
    'sort_hsv_palette_by_cie76_from',
    'sort_hsv_palette_by_cie94_from',
    'sort_hsv_palette_by_ciede2000_from',
    'sort_hsv_palette_by_hsv_euclidean_from',
    'sort_hsl_palette_by_cie76_from',
    'sort_hsl_palette_by_cie94_from',
    'sort_hsl_palette_by_ciede2000_from',
    'sort_hsl_palette_by_hsl_euclidean_from',
    ]

### Palette sorting functions. 

def sort_palette_by_chan(pal, chan):
    """Undocumented"""
    return sorted(pal, key=lambda c: c[chan])


def sort_palette_by_chan0(pal):
    return sorted(pal, key=lambda c: c[0])


# Aliases for sorting channel 0
sort_palette_by_r = sort_palette_by_chan0
sort_palette_by_h = sort_palette_by_chan0


def sort_palette_by_chan1(pal):
    return sorted(pal, key=lambda c: c[1])


# Aliases for sorting channel 1
sort_palette_by_g = sort_palette_by_chan1
sort_palette_by_s = sort_palette_by_chan1


def sort_palette_by_chan2(pal):
    return sorted(pal, key=lambda c: c[2])


# Aliases for sorting channel 2
sort_palette_by_b = sort_palette_by_chan2
sort_palette_by_l = sort_palette_by_chan2
sort_palette_by_v = sort_palette_by_chan2

sort_rgb_palette_by_r = sort_palette_by_r
sort_rgb_palette_by_g = sort_palette_by_g
sort_rgb_palette_by_b = sort_palette_by_b


def sort_rgb_palette_by_hue(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: rgb2hsl(c)[0])


def sort_rgb_palette_by_sat(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: rgb2hsl(c)[1])


def sort_rgb_palette_by_val(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: rgb2hsv(c)[2])


def sort_rgb_palette_by_lum(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: rgb2hsl(c)[2])


sort_hsv_palette_by_hue = sort_palette_by_h
sort_hsv_palette_by_sat = sort_palette_by_s
sort_hsv_palette_by_val = sort_palette_by_v

    
def sort_hsv_palette_by_lum(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsv2hsl(c)[2]) 

    
def sort_hsv_palette_by_r(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsv2rgb(c)[0])

    
def sort_hsv_palette_by_g(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsv2rgb(c)[1])

    
def sort_hsv_palette_by_b(pal):    
    """Undocumented"""
    return sorted(pal, key=lambda c: hsv2rgb(c)[2])


sort_hsl_palette_by_hue = sort_palette_by_h
sort_hsl_palette_by_sat = sort_palette_by_s
sort_hsl_palette_by_lum = sort_palette_by_l


def sort_hsl_palette_by_val(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsl2hsv(c)[2])


def sort_hsl_palette_by_r(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsl2rgb(c)[0])


def sort_hsl_palette_by_g(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsl2rgb(c)[1])


def sort_hsl_palette_by_b(pal):
    """Undocumented"""
    return sorted(pal, key=lambda c: hsl2rgb(c)[2])


def _palette_distances(pal, vec3, distfunc=ciede2000):
    """Undocumented"""
    distances = []
    for i, v in enumerate(pal):
        distances.append((i, distfunc(vec3, v)))
    return distances


def sort_rgb_palette_by_cie76_from(pal, rgb):
    """Undocumented"""
    labpal = rgb2lab(pal)
    lab = rgb2lab(rgb)
    d_list = _palette_distances(labpal, lab, cie76)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_rgb_palette_by_cie94_from(pal, rgb):
    """Undocumented"""
    labpal = rgb2lab(pal)
    lab = rgb2lab(rgb)
    d_list = _palette_distances(labpal, lab, cie94)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_rgb_palette_by_ciede2000_from(pal, rgb):
    """Undocumented"""
    labpal = rgb2lab(pal)
    lab = rgb2lab(rgb)
    d_list = _palette_distances(labpal, lab, ciede2000)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_rgb_palette_by_rgb_euclidean_from(pal, rgb):
    """Undocumented"""
    d_list = _palette_distances(pal, rgb, rgb_euclidean) 
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_rgb_palette_by_hsv_euclidean_from(pal, rgb):
    """Undocumented"""
    hsvpal = rgb2hsv(pal)
    hsv = rgb2hsv(rgb)
    d_list = _palette_distances(hsvpal, hsv, hsv_euclidean) 
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_rgb_palette_by_hsl_euclidean_from(pal, rgb):
    """Undocumented"""
    hslpal = rgb2hsl(pal)
    hsl = rgb2hsl(rgb)
    d_list = _palette_distances(hslpal, hsl, hsl_euclidean) 
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]
    
   
def sort_hsv_palette_by_cie76_from(pal, hsv):
    """Undocumented"""
    labpal = hsv2lab(pal)
    lab = hsv2lab(hsv)
    d_list = _palette_distances(labpal, lab, cie76)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsv_palette_by_cie94_from(pal, hsv):
    """Undocumented"""
    labpal = hsv2lab(pal)
    lab = hsv2lab(hsv)
    d_list = _palette_distances(labpal, lab, cie94)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsv_palette_by_ciede2000_from(pal, hsv):
    """Undocumented"""
    labpal = hsv2lab(pal)
    lab = hsv2lab(hsv)
    d_list = _palette_distances(labpal, lab, ciede2000)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsv_palette_by_hsv_euclidean_from(pal, hsv):
    """Undocumented"""
    d_list = _palette_distances(pal, hsv, hsv_euclidean) 
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsl_palette_by_cie76_from(pal, hsl):
    """Undocumented"""
    labpal = hsl2lab(pal)
    lab = hsl2lab(hsl)
    d_list = _palette_distances(labpal, lab, cie76)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsl_palette_by_cie94_from(pal, hsl):
    """Undocumented"""
    labpal = hsl2lab(pal)
    lab = hsl2lab(hsl)
    d_list = _palette_distances(labpal, lab, cie94)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsl_palette_by_ciede2000_from(pal, hsl):
    """Undocumented"""
    labpal = hsl2lab(pal)
    lab = hsl2lab(hsl)
    d_list = _palette_distances(labpal, lab, ciede2000)
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]


def sort_hsl_palette_by_hsl_euclidean_from(pal, hsl):
    """Undocumented"""
    d_list = _palette_distances(pal, hsl, hsl_euclidean) 
    return [ list(pal[t[0]]) for t in sorted(d_list, key=lambda t: t[1]) ]





"""
The colorways.spectra.* modules provide various approximations of 
the visible spectrum. Each provides a wavelen2rgb(nm) function that
takes a wavelength in namometers and returns an RGB color that is 
the approximate color that light at that wavelength appears to us.
"""




from random import random, normalvariate
from .util import clamp01

"""This module provides functions for bezier curve palettes."""

__all__ = [
    'quadratic_palette',
    'random_quadratic_palette',
    'cubic_palette', 
    'random_cubic_palette', 
]


# Quadratic Bezier 

def _quadratic(x, a, b, c):
    return clamp01(a*(1-x)**2+4*b*x*(1-x)+c*x**2)

def random_quadratic_params():
    return ([random(), normalvariate(0.25, 0.5), random()],
            [random(), normalvariate(0.25, 0.5), random()],
            [random(), normalvariate(0.25, 0.5), random()])

def _quadratic_palette(n, rabc, gabc, babc):
    return [[_quadratic(x/n, *rabc), 
             _quadratic(x/n, *gabc),
             _quadratic(x/n, *babc)] for x in range(n)]

def random_quadratic_palette(n):
    """Returns a random quadratic bezier palette with n colors."""
    return _quadratic_palette(n, *random_quadratic_params())

def quadratic_palette(n, a, b, c):
    """Returns a quadratic bezier palette with n colors defined by points 
       a, b, and c."""
    return _quadratic_palette(n,
        [a[0],b[0],c[0]],
        [a[1],b[1],c[1]],
        [a[2],b[2],c[2]])


# Cubic Bezier

def _cubic(t, a, b, c, d):
    return clamp01(a*(1-t)**3 + 3*b*t*(1-t)**2 + 3*c*t**2*(1-t) + d*t**2)

def random_cubic_params():
    return ([random(), normalvariate(0.25, 0.5), random(), random()],
            [random(), normalvariate(0.25, 0.5), random(), random()],
            [random(), normalvariate(0.25, 0.5), random(), random()])

def _cubic_palette(n, rabcd, gabcd, babcd):
    return [[_cubic(x/n, *rabcd), 
             _cubic(x/n, *gabcd),
             _cubic(x/n, *babcd)] for x in range(n)]

def random_cubic_palette(n):
    """Returns a random cubic bezier palette with n colors."""
    return _cubic_palette(n, *random_cubic_params())

def cubic_palette(n, a, b, c, d):
    """Returns a cubic bezier palette with n colors defined by points 
       a, b, c, and d."""
    return _cubic_palette(n,
        [a[0],b[0],c[0],d[0]],
        [a[1],b[1],c[1],d[1]],
        [a[2],b[2],c[2],d[2]])




from math import sqrt, cos, pi
from .util import clamp01
from .conv import norm2bytes, bytes2norm
from .internals import blend_function

"""This module provides functions for blending colors."""

__all__ = [
    'blend_multiply',
    'blend_screen',
    'blend_overlay',
    'blend_softlight',
    'blend_softlight_ps',
    'blend_softlight_peg',
    'blend_softlight_ill',
    'blend_divide', 
    'blend_lighten',
    'blend_darken',
    'blend_difference',
    'blend_subtract',
    'blend_plus_darker',
    'blend_plus_lighter',
    'blend_linear_dodge',
    'blend_add',
    'blend_average',
    'blend_negate',
    'blend_exclude',
    'blend_hardlight',
    'blend_colorburn',
    'blend_softburn',
    'blend_colordodge',
    'blend_softdodge',
    'blend_reflect',
    'blend_freeze',
    'blend_xor_rgb8',
    'blend_xor',
    'blend_and_rgb8',
    'blend_and',
    'blend_or_rgb8',
    'blend_or',
    'blend_stamp',
    'blend_cosine'
]


# Most formulas from 
# https://en.wikipedia.org/wiki/Blend_modes

@blend_function
def blend_multiply(base, blend):
    """multiply"""
    return [min(1.0, a*b) for a, b in zip(base, blend)]



@blend_function
def blend_screen(base, blend):
    """screen"""
    return [1-(1-a)*(1-b) for a, b in zip(base, blend)]



@blend_function
def blend_overlay(base, blend):
    """overlay"""
    return [2*a*b if a < 0.5 else 1-2*(1-a)*(1-b) for a, b in zip(base, blend)]



@blend_function
def blend_divide(base, blend):
    """Undocumented."""
    return [
        a/b if 0<a<b else 0.0 if a == 0 else 1.0 for a, b in zip(base, blend)]



#@blend_function
#def blend_divide_alt(base, blend):
#    """Undocumented. TODO: based on byte representation"""
#    return [ min(1, a/ ((b+1/256)/256) ) for a, b in zip(base, blend) ]



# W3C Softlight
def _gw3c(a):
    """Undocumented."""
    return ((16*a-12)*a+4)*a if a <= 0.25 else sqrt(a)

def _sftlt_w3c(a, b):
    """Undocumented."""
    if b <= 0.5:
        return a-(1-2*b)*a*(1-a)
    else:   
        return a+(2*b-1)*(_gw3c(a)-a)

@blend_function
def blend_softlight(base, blend):
    """Undocumented."""
    return [_sftlt_w3c(a,b) for a, b in zip(base, blend)] 



# Photoshop Softlight
def _sftlt_ps(a, b):
    """Undocumented."""
    if b < 0.5:
        return 2*a*b+a*a*(1-2*b)
    else:
        return 2*a*(1-b)+sqrt(a)*(2*b-1)

@blend_function
def blend_softlight_ps(base, blend):
    """Undocumented."""
    return [_sftlt_ps(a,b) for a, b in zip(base, blend)] 



# Pegtop Softlight 
# http://www.pegtop.net/delphi/articles/blendmodes/softlight.htm
def _sftlt_peg(a, b):
    """Undocumented."""
    return (1-2*b)*a*a + 2*b*a

@blend_function
def blend_softlight_peg(base, blend):
    """Undocumented."""
    return [_sftlt_peg(a,b) for a, b in zip(base, blend)] 




def _sftlt_ill(a, b):
    """Undocumented."""
    return a**(2**(2*0.5-b))

@blend_function
def blend_softlight_ill(base, blend):
    """Undocumented."""
    return [_sftlt_ill(a,b) for a, b in zip(base, blend)] 




@blend_function
def blend_add(base, blend):
    """Undocumented."""
    return [min(1.0, a+b) for a, b in zip(base, blend)]

blend_linear_dodge = blend_function(blend_add)
blend_plus_lighter = blend_function(blend_add)




@blend_function
def blend_plus_darker(base, blend):
    """Undocumented."""
    return [max(0.0, a+b-1) for a, b in zip(base, blend)]



@blend_function
def blend_subtract(base, blend):
    """Returns base - blend for each channel, clamping negative values to 0."""
    return [max(0.0, a-b) for a, b in zip(base, blend)]



@blend_function
def blend_difference(base, blend):
    """Returns the absolute value of base - blend for each channel."""
    return [abs(a-b) for a, b in zip(base, blend)]



@blend_function
def blend_darken(base, blend):
    """Returns the minimum value for each channel."""
    return [min(a,b) for a, b in zip(base, blend)]



@blend_function
def blend_lighten(base, blend):
    """Returns the mximum value for each channel."""
    return [max(a, b) for a, b in zip(base, blend)]



@blend_function
def blend_average(base, blend):
    """Returns the average value for each channel."""
    return [(a+b)/2.0 for a, b in zip(base, blend)]


#### The following come mostly from 
# http://www.pegtop.net/delphi/articles/blendmodes
# (some modifications may be included.)
@blend_function
def blend_negate(base, blend):
    """
    f(a, b) = 1-abs(1-a-b)
    """
    return [1-abs(1-a-b) for a, b in zip(base, blend)]



@blend_function
def blend_exclude(base, blend):
    """
    f(a,b) = a + b - 2ab
    """
    return [a+b-2*a*b for a, b in zip(base, blend)]


@blend_function
def blend_hardlight(base, blend):
    """
    f(a,b) =  2ab                     (b < 0.5)
              1 - 2(1 - a)(1 - b)     (b >= 0.5)
    """
    return [2*a*b if b < 0.5 else 1-2*(1-a)*(1-b) for a, b in zip(base, blend)]


@blend_function
def blend_colorburn(base, blend):
    """
    f(a,b) = max(0, 1 - (1 - a) / b)
    """
    return [ max(0,1-(1-a)/b) for a, b in zip(base, blend) ]


@blend_function
def blend_softburn(base, blend):
    """
    f(a,b) = 0.5 b / (1 - a)    (a + b < 1)
             1 - 0.5(1 - a) / b (a + b >= 1)
    """
    return [ clamp01(0.5*b/(1-a) if (a+b<1) else 1-0.5*(1-a)/b)
                for a, b in zip(base, blend)]

@blend_function
def blend_colordodge(base, blend):
    """
    f(a,b) = min(1.0, a / (1 - b))
    """
    return [ min(1.0, a/(1-b)) for a, b in zip(base, blend) ]

@blend_function
def blend_softdodge(base, blend):
    """
    f(a,b) =  0.5a / (1 - b) (for a + b < 1)
             1 - 0.5(1 - b) / a (else)
    """
    return blend_softburn(blend, base)

@blend_function
def blend_reflect(base, blend):
    """
    f(a,b) = a**2 / (1 - b)
    """
    return [ min(1.0, a**2/(1-b)) for a, b in zip(base, blend) ]

@blend_function
def blend_freeze(base, blend):
    """
    f(a,b) = 1 - (1 - a)**2 / b
    """
    return [ max(0.0, 1-(1-a)**2/b) for a, b in zip(base, blend) ]

@blend_function
def blend_stamp(base, blend):
    """
    f(a,b) = a + 2b - 1
    """
    return [clamp01( a + 2*b - 1 ) for a, b in zip(base, blend) ]

@blend_function
def blend_cosine(base, blend):
    """
    Was called "interpolate" in Pegtop.
    f(a,b) = ½ - ¼cos(pi*a) - ¼cos(pi*b)
    """
    return [ clamp01( 0.5 - 0.25*cos(pi*a) - 0.25*cos(pi*b) ) 
             for a, b in zip(base, blend)
           ]

@blend_function
def blend_xor_rgb8(base, blend):
    """
    Only intended for use with colors in rgb8. 
    """
    return[ a ^ b for a, b in zip(base, blend) ]

@blend_function
def blend_xor(base, blend):
    return bytes2norm(blend_xor_rgb8(norm2bytes(base),norm2bytes(blend)))

@blend_function
def blend_and_rgb8(base, blend):
    """
    Only intended for use with colors in rgb8. 
    """
    return[ a & b for a, b in zip(base, blend) ]

@blend_function
def blend_and(base, blend):
    return bytes2norm(blend_and_rgb8(norm2bytes(base),norm2bytes(blend)))

@blend_function
def blend_or_rgb8(base, blend):
    """
    Only intended for use with colors in rgb8. 
    """
    return[ a | b for a, b in zip(base, blend) ]

@blend_function
def blend_or(base, blend):
    return bytes2norm(blend_or_rgb8(norm2bytes(base),norm2bytes(blend)))



from random import random
from math import tau, cos, sin, hypot, atan2
"""
This module provides miscellaneous utility functions.

Exports:
    Functions for range limiting:
        reflect(x)
        clamp01(x)

    Validation functions:
        is_vec(arg, check_min=None, check_max=None, check_len=None, 
               min_len=None, max_len=None, ok_types=(float,int))
        is_pal01(arg)
        is_pal(arg)
        is_byte_pal(arg)

    Linear interpolation utils: 
        lspace(u, v, n=2)
        circ_lspace(a, b, n=2, bigarc=False)
        hxx_lspace(u, v, n, bigarc=False)

    Palette modification utils:
        apply_to_chans(vec3, chans, fn, *args, **kwargs)
        modify_palette(pal, chans, fn, *args, **kwargs)
        pal_extract_chans(pal):

    Vec3 utils:
        randvec3()
        randvec3mono()
        randvec3_partial(vec3)
        mix_vec3(base, mixin, weight)
        randmix_vec3(base, weight)

    Coordinate conversions:
        rect(r, theta, unit='rad')
        polar(x, y, unit='rad')

    Miscellaneous:
        circ_mean(angles, unit='rad')
        luma(rgb, use='y601'): 
"""

__all__ = [
    'reflect', 'clamp01',
    'is_vec', 'is_pal01', 'is_pal', 'is_byte_pal',
    'lspace', 'circ_lspace', 'hxx_lspace', 'fspace',
    'apply_to_chans', 'modify_palette', 'pal_extract_chans',
    'randvec3', 'randvec3mono', 'randvec3_partial', 'mix_vec3', 'randmix_vec3',
    'rect', 'polar',
    'circ_mean', 'luma',
    ]


### General utilities 

def reflect(x):
    """
    Maps x onto [0,1], treating x as a distance to "walk"
    back and forth across the interval starting at 0.
    """
    r = abs(x) % 2
    return r if r < 1 else 2-r


def clamp01(x):
    """
    Returns 0 if x is less than 0, 1 if x is greater than 1, and x otherwise.
    """
    return min(max(0.0, x), 1.0)


### Validation utilities.

def is_vec(arg, check_min=None, check_max=None, check_len=None,
           min_len=None, max_len=None, ok_types=(float,int)):
    if type(arg) is not list: return False
    if check_len is not None and check_len != len(arg): return False
    if min_len is not None and min_len > len(arg): return False
    if max_len is not None and max_len < len(arg): return False
    if not all( type(e) in ok_types for e in arg ): return False
    if check_min is not None and check_min > min(arg): return False
    if check_max is not None and check_max < max(arg): return False
    return True

def is_pal01(arg):
    if type(arg) is not list: return False
    return all(
        is_vec(e, min_len=3, check_min=0, check_max=1) for e in arg)

def is_pal(arg):
    if type(arg) is not list: return False
    return all( is_vec(e, min_len=3) for e in arg)

def is_byte_pal(arg, require_ints=False):
    if type(arg) is not list: return False
    if require_ints:
        return all( is_vec(
                e, min_len=3, check_min=0, check_max=255, ok_types=(int,)
            ) for e in arg)
    else:
        return all( is_vec(
                e, min_len=3, check_min=0, check_max=255
            ) for e in arg)

      


# Linear interpolation utils

def _lerp(a, b, t):
    """Undocumented"""
    return a + t*(b-a)



def lspace(u, v, n=2):
    """Undocumented"""
    if n < 2: n = 2
    ret = []
    if type(u) == type(v) == list:
        dims = min(len(u), len(v))
        for i in range(n): 
            ret.append([u[d] + i/(n-1)*(v[d]-u[d]) for d in range(dims)])
        return ret
    else:
        return lspace([u], [v], n)

def fspace(u, v, n=2, fn=None):
    """Undocumented"""
    if fn is None:
        fn = lambda x: x
    if n < 2: n = 2
    return [ fn(e) for e in lspace(u, v, n) ]


def circ_lspace(a, b, n, bigarc=False):
    """Undocumented"""
    d1 = abs(a-b)
    d2 = 1 - abs(a-b)
    if (0 < d2 < d1 and not bigarc) or (0 < d1 < d2 and bigarc):
        deltas = [e[0] for e in lspace([0], [d2], n)]
        if a < b: 
            return [(1 + (a-d))%1 for d in deltas]
        else: 
            return [(a+d)%1 for d in deltas]
    else: 
        return [e[0] for e in lspace([a], [b], n)]



def hxx_lspace(u, v, n, bigarc=False): 
    h_chan = circ_lspace(u[0], v[0], n, bigarc)
    xx_chans = lspace(u[1:], v[1:], n)
    return [ [h_chan[i]] + xx for i, xx in enumerate(xx_chans) ]
    
   


# Palette modification


def apply_to_chans(vec3, chans, fn, *args, **kwargs):
    """Undocumented"""
    new = []
    if type(chans) is int:
        chans = [chans]
    L = len(vec3) # because maybe it isn't a vec3.
    for i in range(L):
        if i in chans:
            new.append( fn(vec3[i], *args, **kwargs))
        else:
            new.append(vec3[i])
    return new


def modify_palette(pal, chans, fn, *args, **kwargs):
    """Undocumented"""
    new = []
    n = len(pal)
    for i, v in enumerate(pal): 
        new.append(apply_to_chans(v, chans, fn, i, n, *args, **kwargs))
    return new 

def pal_extract_chans(pal):
    """
    Takes a palette (list of vec3s) and returns a tuple of 3 lists, each 
    containing the values from one of the 3 vec3 channels. 
    E.g. pal_extract_chans([ [1,2,3], [4,5,6] ]) returns ([1,4], [2,5], [3,6])
    """
    chan0 = []
    chan1 = []
    chan2 = []
    for v in pal:
        chan0.append(v[0])
        chan1.append(v[1])
        chan2.append(v[2])
    return (chan0, chan1, chan2)


# Vec3 utils

def _rand_norm_byte():
    return 255*random()/255.0

def random_color():
    return [ _rand_norm_byte() for _ in range(3) ]

def random_gray():
    return [ _rand_norm_byte() ] * 3

def rand_color_partial(vec3):
    return [ rand_color() if c is None else c for c in vec3 ]

def randvec3():
    """Returns a list of three random floats in [0,1]."""
    return [ random() for _ in range(3) ]

def randvec3mono():
    """Returns a list containing a random float in [0,1] repeated 3 times."""
    return [ random() ] * 3   

def randvec3_partial(vec3):
    """
    Returns a copy of the given list with None elements replaced with random
    floats in the range [0,1].
    """
    return [ random() if c is None else c for c in vec3 ]

def mix_vec3(base, mixin, weight):
    """
    Takes a base vec3, a mixin vec3, and a weight and returns a vec3 whose 
    components are a weighted average of the given vec3s (weighted towards 
    the base.)
    """
    return [ weight*base[i] + (1-weight)*mixin[i] for i in range(len(base)) ]



def randmix_vec3(base, weight):
    """
    Takes a base vec3 and a weight and returns a vec3 whose components are the 
    weighted average of the base and a random vec3 (weighted towards the base.)
    """
    return mix_vec3(base, randvec3(), weight)



# Coordinate conversion

def rect(r, theta, unit='rad'):
    """
    Convert from polar to cartesian coordinates.
    The 'unit' keyword argument can take the values
    'turn', 'deg', or 'rad' and specifies the units
    that theta is given in.
    """
    if unit == 'turn': 
        theta *= tau
    elif unit == 'deg':
        theta = theta/360 * tau 
    return r*cos(theta), r*sin(theta)


def polar(x, y, unit='rad'):
    """
    Convert from cartesian to polar coordinates.
    The 'unit' keyword argument can take the values
    'turn', 'deg', or 'rad'. 
    """
    r = hypot(x, y)
    theta = atan2(y, x)
    if unit == 'turn':
        theta = theta/tau
    elif unit == 'deg':
        theta = theta * 360/tau
    return r, theta


# Miscellaneous

def circ_mean(angles, unit='rad'):
    # TODO: unit is ignored.
    """
    Returns the circular mean of the given angles in turns.
    TODO: unit is ignored.
    """
    xs, ys = [], []
    n = len(angles)
    for a in angles: 
        x, y = rect(1.0, a, unit='turn') 
        xs.append(x)
        ys.append(y)
    return polar(sum(xs)/n, sum(ys)/n, unit='turn')


def luma(rgb, use='y601'): 
    """
    Returns the luma value of the given rgb color using the specified set of 
    coefficients (default: y601, available: y601 y240, y709, y2020)
    """
    # coefficients from https://en.wikipedia.org/wiki/HSL_and_HSV
    y601  = (0.2989, 0.5870, 0.1140) # NTSC SDTV
    y240  = (0.2120, 0.7010, 0.0870) # Adobe RGB
    y709  = (0.2126, 0.7152, 0.0722) # HDTV
    y2020 = (0.2627, 0.6780, 0.0593) # UHDTV, HDR
    coeffs = y601
    if use == 'y240': coeffs = y240
    if use == 'y709': coeffs = y709
    if use == 'y2020': coeffs = y2020
    return (rgb[0]*coeffs[0] + rgb[1]*coeffs[1] + rgb[2]*coeffs[2])




from ..conv import hsl_to_hsv

"""This module contains utilities for Processing's Python mode."""

###  Functions in this module require the Processing environment.

def hsv_from_color(c):
    """Get a list of the HSV/HSB components from a processing color.

    Args:
        A Processing color object.

    Returns:
        A 3-element list (vec3) of the color argument's hue, 
        saturation, and value (i.e. brightness) components.
    """
    return [hue(c), saturation(c), brightness(c)]



def rgb_from_color(c):
    """Get a list of the RGB components from a processing color.
    
    Args:
        c: a Processing color object.

    Returns: 
        A 3-element list (vec3) of the color argument's red, green, 
        and blue components. 
    """
    return [red(c), green(c), blue(c)]



def get_color(palette, i, mode=None):
    """Retrieves a palette color as a Processing color object.

    Args:
        palette: a list of vec3 colors in HSL. 
        i: an index into the palette
        mode: a Processing color mode (RGB or HSB.) Default is HSB.

    Returns:
        A Processing color object.
    """
    if mode is None:
        vec3 = hsl_to_hsv(palette[i])
        return as_color_in_mode(vec3, HSB)
    return as_color_in_mode(palette[i], mode)
  


def as_color_in_mode(vec3, mode):
    """Converts a vec3 to a Processing color object.

    Args:
        vec3: a list containing 3 floats in the range [0,1]
        mode: a Processing color mode (either RGB or HSB)

    Returns:
        A Processing color object.
    """
    push()
    colorMode(mode, 1.0)
    c = color(*vec3)
    pop()
    return c
   
   
   
def as_hsb_color(vec3):
    """Converts a vec3 to a Processing color object in HSB mode.
    
    Args:
        vec3: a list containing 3 floats in the range [0,1]

    Retruns:
        A Processing color object.
    """ 
    return as_color_in_mode(vec3, HSB)

    
    
def as_rgb_color(vec3):
    """Converts a vec3 to a Processing color object in RGB mode.
    
    Args:
        vec3: a list containing 3 floats in the range [0,1]

    Retruns:
        A Processing color object.
    """
    return as_color_in_mode(vec3, RGB)






#from math import tau, cos, sin, hypot, atan2
from statistics import mean, pstdev, pvariance
#from .conv import * 
from .dist import cie76
from .conv import (
    rgb2hsl,
    rgb2hsv,
    hsv2hwb,
    rgb2lab,
    hsv2hsl,
    hsv2rgb,
    hsl2hsv,
    hsl2rgb,
    hwb2hsv,
    hex2lab,
    norm2bytes,
    bytes2norm,
)
from .util import (
    pal_extract_chans, rect, polar, circ_mean, luma,
)

"""Provides the MasterPalette class and utility functions.

"""
__all__ = [
    'new_order_by_dist_closure'
    'MasterPalette'
]


def list_indexed_by(L, indices):
    return [L[i] for i in indices]

def new_order_by_dist_closure(hexstr, fn=cie76): 
    lab = hex2lab(hexstr)
    def orderfunc(p, i): 
        return fn(p.lab[i], lab)
    return orderfunc



def bin_sizes(nitems, nbins):
    bsz = nitems / nbins
    return [round((i+1)*bsz)-round(i*bsz) for i in range(nbins)]

def bin_slices(nitems, nbins):
    bsz = nitems/nbins
    return [(round(i*bsz), round((i+1)*bsz)) for i in range(nbins)]




class MasterPalette: 
    """
    A MasterPalette holds palette data and metadata and provides 
    methods to facilitate palette analysis tasks, the creation and 
    storage of different palette orderings, and the generation of 
    palette variants. 
    """
    
    def __init__(self, palette, mode='hsl', usebytes=False):
        self._palette = list(palette) # Keep a copy of the original.
        self._mode = mode
        self._usebytes = usebytes

        self.pal = self._palette
        self.curr = self.pal

        if usebytes: 
            self.pal = bytes2norm(palette)
    
        if mode == 'rgb':
            self.rgb = self.pal
            self.hsl = rgb2hsl(self.rgb)
            self.hsv = rgb2hsv(self.rgb)
            self.hwb = hsv2hwb(self.hsv) # preferred
            self.lab = rgb2lab(self.rgb)
        elif mode == 'hsv':
            self.hsv = sel.pal
            self.hsl = hsv2hsl(self.hsv)
            self.hwb = hsv2hwb(self.hsv)
            self.rgb = hsv2rgb(self.hsv)
            self.lab = rgb2lab(self.rgb)
        elif mode == 'hsl':
            self.hsl = self.pal
            self.hsv = hsl2hsv(self.hsl)
            self.hwb = hsv2hwb(self.hsv) # preferred
            self.rgb = hsl2rgb(self.hsl)
            self.lab = rgb2lab(self.rgb)
        elif mode == 'hwb':
            self.hwb = self.pal
            self.hsv = hwb2hsv(self.hwb)
            self.hsl = hsv2hsl(self.hsv) 
            self.rgb = hsl2rgb(self.hsl) # preferred 
            self.lab = rgb2lab(self.rgb)

        self.rgb8 = norm2bytes(self.rgb)
    
        self.hsl_h, self.hsl_s, self.hsl_l = pal_extract_chans(self.hsl)
        self.hsv_h, self.hsv_s, self.hsv_v = pal_extract_chans(self.hsv)
        self.rgb_r, self.rgb_g, self.rgb_b = pal_extract_chans(self.rgb)
        self.lab_l, self.lab_a, self.lab_b = pal_extract_chans(self.lab)
        self.hwb_h, self.hwb_w, self.hwb_b = pal_extract_chans(self.hwb)

        self.luma = [ luma(clr) for clr in self.rgb ]
        self.n = len(self.hsl)

        self.mean_hue_dist, self.mean_hue = circ_mean(self.hsl_h)
        if self.mean_hue < 0.0: 
            self.mean_hue += 1.0
        
        self.hsl_mean_s = mean(self.hsl_s)
        self.hsl_std_s = pstdev(self.hsl_s)
        self.hsl_min_s = min(self.hsl_s)
        self.hsl_max_s = max(self.hsl_s)
        self.hsl_mean_l = mean(self.hsl_l)
        self.hsl_std_l = pstdev(self.hsl_l)
        self.hsl_min_l = min(self.hsl_l)
        self.hsl_max_l = max(self.hsl_l)

        self.hsv_mean_s = mean(self.hsv_s)
        self.hsv_std_s = pstdev(self.hsv_s)
        self.hsv_min_s = min(self.hsv_s)
        self.hsv_max_s = max(self.hsv_s)
        self.hsv_mean_v = mean(self.hsv_v)
        self.hsv_std_v = pstdev(self.hsv_v)
        self.hsv_min_v = min(self.hsv_v)
        self.hsv_max_v = max(self.hsv_v)

        self.rgb_mean_r = mean(self.rgb_r)
        self.rgb_std_r = pstdev(self.rgb_r)
        self.rgb_min_r = min(self.rgb_r)
        self.rgb_max_r = max(self.rgb_r)
        self.rgb_mean_g = mean(self.rgb_g)
        self.rgb_std_g = pstdev(self.rgb_g)
        self.rgb_min_g = min(self.rgb_g)
        self.rgb_max_g = max(self.rgb_g)
        self.rgb_mean_b = mean(self.rgb_b)
        self.rgb_std_b = pstdev(self.rgb_b)
        self.rgb_min_b = min(self.rgb_b)
        self.rgb_max_b = max(self.rgb_b)

        self.hwb_mean_w = mean(self.hwb_w)
        self.hwb_std_w = pstdev(self.hwb_w)
        self.hwb_min_w = min(self.hwb_w)
        self.hwb_max_w = max(self.hwb_w)
        self.hwb_mean_b = mean(self.hwb_b)
        self.hwb_std_b = pstdev(self.hwb_b)
        self.hwb_min_b = min(self.hwb_b)
        self.hwb_max_b = max(self.hwb_b)

        self.luma_mean = mean(self.luma)
        self.luma_std = pstdev(self.luma)
        self.luma_min = min(self.luma)
        self.luma_max = max(self.luma)
  
        indexes = range(self.n)
        self.indexes = indexes

        self.orderings = {
            None: indexes,
            'by-hue': sorted(indexes, key=lambda i: self.hsl_h[i]), 
            'by-lsat': sorted(indexes, key=lambda i: self.hsl_s[i]), 
            'by-lum': sorted(indexes, key=lambda i: self.hsl_l[i]), 
            'by-vsat': sorted(indexes, key=lambda i: self.hsv_s[i]), 
            'by-val': sorted(indexes, key=lambda i: self.hsv_v[i]), 
            'by-r': sorted(indexes, key=lambda i: self.rgb_r[i]), 
            'by-g': sorted(indexes, key=lambda i: self.rgb_g[i]), 
            'by-b': sorted(indexes, key=lambda i: self.rgb_b[i]), 
            'by-luma': sorted(indexes, key=lambda i: self.luma[i]), 
            'by-L': sorted(indexes, key=lambda i: self.lab_l[i]), 
            'by-white': sorted(indexes, key=lambda i: self.hwb_w[i]), 
            'by-black': sorted(indexes, key=lambda i: self.hwb_b[i]), 
        }

    def get_rgb(self, order=None):
        return list_indexed_by(self.rgb, self.orderings[order])

    def get_hsl(self, order=None):
        return list_indexed_by(self.hsl, self.orderings[order])

    def get_hsv(self, order=None):
        return list_indexed_by(self.hsl, self.orderings[order])

    def get_lab(self, order=None):
        return list_indexed_by(self.hsl, self.orderings[order])

    def get_hwb(self, order=None):
        return list_indexed_by(self.hwb, self.orderings[order])

    def get_rgb8(self, order=None):
        return list_indexed_by(self.rgb8, self.orderings[order])

    def register_ordering(self, name, ordering):
        self.orderings[name] = ordering

    def create_ordering(self, name, order_by_fn):
        idxkey = []
        # Use provided function to compute key for index.
        for i in range(self.n):
            idxkey.append((i, order_by_fn(self, i)))
        ordering = [t[0] for t in sorted(idxkey, key=lambda t: t[1])]
        self.register_ordering(name, ordering)

    def create_bin_ordering(self, name, bins, base_order, bin_order):
        """
        Creates a new bin ordering and saves it. 

        First, the palette is ordered according to the base_order, then it 
        is partitioned into bins. Finally, the contents of each bin are sorted 
        according to the bin_order.

        Args:
            name (str): the name to save the ordering as.
            bins (int): the number of bins.
            base_order (str): the name of the ordering to start with.
            bin_order (str): the name of the ordering to use for bin contents.
        """
        order = list(self.orderings[base_order]) # Start with base order.
        bslices = bin_slices(self.n, bins)
        # create a lookup table for the ordering given by key2. 
        lookup = [None] * self.n
        for i, ix in enumerate(self.orderings[bin_order]): 
            lookup[ix] = i

        # Iterate through the slices and sort each slice by the ordering
        # specified with key2. 
        for s in bslices:
            order[s[0]:s[1]] = sorted(order[s[0]:s[1]], key=lambda i: lookup[i])

        self.register_ordering(name, order) 

    def gen(self, fn):
        """
        Generate a variant.

        This method returns a list.  The provided callback is called once for 
        each color in the MasterPalette. The callback must have a signature 
        such as:

            callback(master, i)

        It must accept two arguments; the first will be a reference to the 
        MasterPalette and the second is an index that increases by one on each 
        call, ranging from 0 on the first call to one less than the length of 
        the MasterPalette on the final call. 
        
        When the callback returns None, nothing is added to the return list.
        When it returns a set or a tuple, each element in the set or tuple is 
        added to the return list. Otherwise, whatever the callback returns is 
        added to the return list. 

        Most often, the values to be added to the return list will be three
        element lists (vec3's) so gen will return a palette. But this is not
        strictly required. No checking is done on the elements added to the
        return list.
        """

        new = []
        for i in range(self.n):
            retval = fn(self, i)
            if retval is not None:
                if type(retval) in (set, tuple):
                    for vec3 in retval: 
                        new.append(vec3)
                else:
                    new.append(retval)
        return new        
        
    def __len__(self):
        return self.n

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self.curr[i])
        else:
            return self.curr[i]

    def reset(self):
        self.curr = self.pal

    def config(self, space=None, order=None):
        if space is None: 
           space = self._mode
        if space == 'rgb':
            self.curr = self.get_rgb(order) 
        elif space == 'hsl':
            self.curr = self.get_hsl(order) 
        elif space == 'hsv':
            self.curr = self.get_hsv(order) 
        elif space == 'lab':
            self.curr = self.get_lab(order) 
        elif space == 'hwb':
            self.curr = self.get_hwb(order) 
        elif space == 'rgb8':
            self.curr = self.get_rgb8(order) 
        else: 
            self.curr = list_indexed_by(self.pal, self.orderings[order])

    def set_variant(self, fn):
        self.curr - self.gen(fn)



from functools import wraps

"""Internally used functions."""

__all__ = [
    'blend_function',
    'colpal_function',
    'colpal_params_function',
    'hexpal_function',
]


# Handles functions with two arguments: either two colors, 
# a palette and a color, or two palettes. 
def blend_function(_function): 
    @wraps(_function)
    def blend_closure(base, blend):
        if type(base[0]) is list: 
            if type(blend[0]) is list: 
                return [ _function(a, b) for a, b in zip(base, blend) ]
            else:
                return [_function(a, blend) for a in base ]
        else:
            return _function(base, blend)
    return blend_closure


# Handles functions with one argument which may be a vec3 or palette. 
def colpal_function(_function):
    @wraps(_function)
    def cp_closure(base):
        if type(base[0]) is list:
            return [ _function(a) for a in base ] 
        else:
            return _function(base)                    
    return cp_closure


# Handles functions with a first argument that may be a vec3 or a palette. 
def colpal_params_function(_function):
    @wraps(_function)
    def cp_params_closure(base, *args, **kwargs):
        if type(base[0]) is list:
            return [ _function(a, *args, **kwargs) for a in base ] 
        else:
            return _function(base, *args, **kwargs)                    
    return cp_params_closure


def hexpal_function(_function):
    @wraps(_function)
    def hexpal_closure(hexpal):
        if type(hexpal) is str and 5<len(hexpal)<8:
            return _function(hexpal)
        elif type(hexpal) is list:
            return [ _function(hexcode) for hexcode in hexpal ]
        else:
            raise ValueError('Expected hexcode or list of hexcodes.')
    return hexpal_closure





"""
This module provides functions defined in the Web Content Accessibility 
Guidelines 2.0. 

Exports:
wcag_rel_lum(rgb): Computes a color's relative luminance.
wcag_contrast_ratio(rgb1, rgb2): Computes contrast ratio between colors.

"""

__all__ = ['wcag_rel_lum', 'wcag_contrast_ratio']


def wcag_rel_lum(rgb):
    """
    Computes the relative luminance (as defined by WCAG 2.0) of the given 
    vec3 interpreted as RGB.
    """
    # Computed according to:
    # https://www.w3.org/TR/WCAG20/#relativeluminancedef
    # This uses a different threshold value for each color channel (0.03928) 
    # than is used for the sRGB to XYZ conversion done by rgb2xyz(). 
    # (Relative luminance is the Y in XYZ.)
    #
    # if Cr <= 0.03928 then R = Cr/12.92 else R = ((Cr+0.055)/1.055) ^ 2.4
    trgb = [ c/12.92 if c <= 0.03928 else ((c+0.055)/1.055)**2.4 for c in rgb ]
    #L = 0.2126 * R + 0.7152 * G + 0.0722 * B
    return (0.2126 * trgb[0]) + (0.7152 * trgb[1]) + (0.0722 * trgb[2])

   
 
def wcag_contrast_ratio(rgb1, rgb2):
    """
    Computes the contrast ratio (as defined by WCAG 2.0) of the two given 
    vec3s interpreted as RGB values.
    """
    # https://www.w3.org/TR/WCAG20/#contrast-ratiodef
    # (L1 + 0.05) / (L2 + 0.05), where
    # L1 is the relative luminance of the lighter of the colors, and
    # L2 is the relative luminance of the darker of the colors.
    L1 = wcag_rel_lum(rgb1)
    L2 = wcag_rel_lum(rgb2)
    L1, L2 = max(L1, L2), min(L1, L2)
    return (L1 + 0.05) / (L2 + 0.05)






from math import pi, sqrt, sin, cos, exp, atan2

"""
This module provides functions for computing various distance measures 
between colors.

Exports:
    rgb_euclidean(rgb1, rgb2): Computes euclidean distance in RGB space.
    hsv_euclidean(hsv1, hsv2): Computes euclidean distance in HSV space.
    hsl_euclidean(hsl1, hsl2): Computes euclidean distance in HSL space. 
    hwb_euclidean(hsl1, hsl2): Computes euclidean distance in HWB space. 
    cie76(lab1, lab2): 
    cie94(lab1, lab2, kL=1, K1=0.045, K2=0.015):
    ciede2000(lab1, lab2, kL=1.0, kC=1.0, kH=1.0): 
    cie76_JND: The "Just Noticeable Difference" threshold for CIE76 distances.

"""

__all__ = [
    'rgb_euclidean', 
    'hsv_euclidean', 
    'hsl_euclidean',
    'cie76', 
    'cie76_JND', 
    'cie94', 
    'ciede2000',
    ]

tau = 2*pi # Defining because python2's math module doesn't provide it. 

# Distance functions
# TODO: This module needs some clean up. euclidean_rgb() isn't limited to 
#       RGB, so its name is poor. hsv_euclidean and hsl_euclidean are the 
#       same. (Look into HWB version; it's probably the same as well.) 
#       Not sure if a LAB version makes sense. 


def rgb_euclidean(rgb1, rgb2):
    """Returns the euclidean distance between the vec3 args."""
    return sqrt((rgb1[0]-rgb2[0])**2 + (rgb1[1]-rgb2[1])**2 + (rgb1[2]-rgb2[2])**2)


def hxx_euclidean(hxx1, hxx2):
    """Returns euclidean distance between two colors in an Hxx colorspace."""
    x1, y1, z1 = cos(hxx1[0]*tau)*hxx1[1], sin(hxx1[0]*tau)*hxx1[1], hxx1[2]
    x2, y2, z2 = cos(hxx2[0]*tau)*hxx2[1], sin(hxx2[0]*tau)*hxx2[1], hxx2[2]
    return sqrt( (x2-x1)**2 + (y2-y1)**2 + (z2-z1)**2 )

hsv_euclidean = hxx_euclidean
hsl_euclidean = hxx_euclidean
hwb_euclidean = hxx_euclidean



# Computes delta E using CIE 1976
def cie76(lab1, lab2): 
    """Computes the CIE 1976 version of Delta E between two L*a*b* colors."""
    # https://en.wikipedia.org/wiki/Color_difference#CIE76
    return sqrt((lab1[0]-lab2[0])**2 + (lab1[1]-lab2[1])**2 + (lab1[2]-lab2[2])**2)

cie76_just_noticeable_difference = 2.3
cie76_JND = 2.3



# Computes delta E using CIE 1994
def cie94(lab1, lab2, kL=1, K1=0.045, K2=0.015):
    """Computes the CIE 1994 version of Delta E between two L*a*b* colors."""
    # https://en.wikipedia.org/wiki/Color_difference#CIE94
    L1, a1, b1 = lab1 
    c1 = sqrt(a1**2 + b1**2)
    L2, a2, b2 = lab2 
    c2 = sqrt(a2**2 + b2**2)
    dL2 = (L1 - L2)**2
    dC2 = (c1 - c2)**2 
    dH2 = (a1 - a2)**2 + (b1 - b2)**2 - dC2
    
    return sqrt( 
            dL2/kL**2 + 
            dC2/(1 + K1*c1)**2 +
            dH2/(1 + K2*c1)**2)



# Computes delta E using CIE DE2000
def ciede2000(lab1, lab2, kL=1.0, kC=1.0, kH=1.0): 
    """Computes the CIE 2000 version of Delta E between two L*a*b* colors."""
    # Reference, including Gaurav Sharma's original matlab implementation:
    # http://www2.ece.rochester.edu/~gsharma/ciede2000/
    L1, a1, b1 = lab1
    c1 = sqrt( a1**2 + b1**2)

    L2, a2, b2 = lab2
    c2 = sqrt( a2**2 + b2**2)

    c_avg = (c1 + c2) / 2

    G = 0.5 * (1 - sqrt(c_avg**7 / (c_avg**7 + 25**7))) 

    ap1 = a1 * (1 + G)
    ap2 = a2 * (1 + G)

    cp1 = sqrt(ap1**2 + b1**2)
    cp2 = sqrt(ap2**2 + b2**2)   

    hp1 = 0 if abs(ap1) + abs(b1) == 0 else atan2(b1, ap1)
    hp1 = hp1 if hp1 >= 0 else hp1 + tau

    hp2 = 0 if abs(ap2) + abs(b2) == 0 else atan2(b2, ap2)
    hp2 = hp2 if hp2 >= 0 else hp2 + tau

    dL = L2 - L1
    dC = cp2 - cp1
   
    dhp = 0 if cp1*cp2 == 0 else hp2 - hp1 
    if dhp >  pi: dhp -= tau
    if dhp < -pi: dhp += tau

    dH = 2 * sqrt(cp1 * cp2) * sin(dhp / 2)

    Lp = (L1 + L2) / 2
    Cp = (cp1 + cp2) / 2

    hp =  hp1 + hp2

    if cp1 * cp2 != 0: 
        hp = (hp1 + hp2) / 2.0
        if abs(hp1 - hp2) > pi:
            hp -= pi 
        if hp < 0: 
            hp += tau

    Lpm = (Lp - 50)**2
    T = (1 - 0.17 * cos(hp - tau / 12) + 
        0.24 * cos(2 * hp) + 
        0.32 * cos(3 * hp + tau / 60) -
        0.20 * cos(4 * hp - 63 * tau / 360))
    
    Sl = 1 + (0.015 * Lpm) / sqrt(20 + Lpm)
    Sc = 1 + 0.045 * Cp
    Sh = 1 + 0.015 * Cp * T
    
    deltaTheta = 30 * tau/360 * exp(-1 * ((360/tau * hp - 275)/25)**2)
    Rc = 2 * sqrt( Cp**7 / (Cp**7+25**7))

    Rt = -1 * sin(2 * deltaTheta) * Rc

    return sqrt(
        (dL / (kL * Sl))**2 + 
        (dC / (kC * Sc))**2 + 
        (dH / (kH * Sh))**2 + 
        Rt * dC / (kC * Sc) * dH / (kH * Sh))





from .internals import colpal_function
from .conv import (
    rgb2hsl, rgb2hsv, rgb2hwb, rgb2lab, 
    hsl2rgb, hsv2rgb, hwb2rgb, lab2rgb )

__all__ = [
    'ryb2rgb',
    'ryb2hsl',
    'ryb2hsv',
    'ryb2hwb',
    'ryb2lab',
    'rgb2ryb',
    'hsl2ryb',
    'hsv2ryb',
    'hwb2ryb',
    'lab2ryb',
]

"""
This module provides RYB color space conversions. 
"""

################################################################################
# The functions rgb2ryb and ryb2rgb contain only minor modification to the 
# original versions by Arah J. Leonard found here:
# http://web.archive.org/web/20130415143903/http://www.insanit.net/tag/rgb-to-ryb/
#
# The original copyright notice and licensing: 
# Author: Arah J. Leonard
# Copyright 01AUG09
# Distributed under the LGPL - http://www.gnu.org/copyleft/lesser.html
# ALSO distributed under the The MIT License from the Open Source Initiative 
# (OSI) - http://www.opensource.org/licenses/mit-license.php
# You may use EITHER of these licenses to work with / distribute this source 
# code.
# Enjoy!


# Convert a red-green-blue system to a red-yellow-blue system.
@colpal_function
def rgb2ryb(rgb):
    r, g, b = rgb

    # Remove the whiteness from the color.
    w = min(r, g, b)
    r = r - w
    g = g - w
    b = b - w

    mg = max(r, g, b)

    # Get the yellow out of the red+green.
    y = min(r, g)
    r -= y
    g -= y

    # If this unfortunate conversion combines blue and green, 
    # then cut each in half to preserve the value's maximum range.
    if b and g:
        b /= 2.0
        g /= 2.0

    # Redistribute the remaining green.
    y += g
    b += g

    # Normalize to values.
    my = float(max(r, y, b))
    if my:
        n = mg / my
        r *= n
        y *= n
        b *= n

    # Add the white back in.
    r += w
    y += w
    b += w

    # And return back the ryb typed accordingly.
    #if t is int:
    #    print("rounding")
    #    return[round(r), round(g), round(b)]

    #return [t(r), t(y), t(b)]
    return [r, y, b]



# Convert a red-yellow-blue system to a red-green-blue system.
@colpal_function
def ryb2rgb(ryb):
    r, y, b = ryb

    # Remove the whiteness from the color.
    w = (min(r, y, b))
    r = r - w
    y = y - w
    b = b - w

    my = max(r, y, b)

    # Get the green out of the yellow and blue
    g = min(y, b)
    y -= g
    b -= g

    if b and g:
        b *= 2.0
        g *= 2.0

    # Redistribute the remaining yellow.
    r += y
    g += y

    # Normalize to values.
    mg = float(max(r, g, b))
    if mg:
        n = my / mg
        r *= n
        g *= n
        b *= n

    # Add the white back in.
    r += w
    g += w
    b += w

    return [r, g, b]


def ryb2hsl(ryb):
    return rgb2hsl(ryb2rgb(ryb))

def ryb2hsv(ryb):
    return rgb2hsv(ryb2rgb(ryb))

def ryb2hwb(ryb):
    return rgb2hwb(ryb2rgb(ryb))

def ryb2lab(ryb):
    return rgb2lab(ryb2rgb(ryb))


def hsl2ryb(hsl):
    return rgb2ryb(hsl2rgb(hsl))

def hsv2ryb(hsv):
    return rgb2ryb(hsv2rgb(hsv))

def hwb2ryb(hwb):
    return rgb2ryb(hwb2rgb(hwb))

def lab2ryb(lab):
    return rgb2ryb(lab2rgb(lab))




import colorsys
from .internals import colpal_function, hexpal_function
from .util import clamp01

"""
This module provides conversion functions for colors and palettes.

Exports:
    Color space conversion functions. These take a single argument
    which is expected to be either a 3-element list (i.e. a color)
    or a list of colors (i.e. a palette.) 
    
    hsl2hsv hsl2hwb hsl2rgb hsl2lab 
    hsv2hsl hsv2hwb hsv2rgb hsv2lab 
    hwb2hsl hwb2hsv hwb2rgb hwb2lab 
    rgb2hsl rgb2hsv rgb2hwb rgb2lab 
    lab2hsl lab2hsv lab2hwb lab2rgb 

    hex2hsl hex2hsv hex2hwb hex2rgb hex2lab 
    hsl2hex hsv2hex hwb2hex rgb2hex lab2hex 

    bytes2norm
    norm2bytes
    hex2bytes
    bytes2hex

    rgb2hue
    hxx_in_degpct

    ]

"""

__all__ = [
    'hsl2hsv', 'hsl2hwb', 'hsl2rgb', 'hsl2lab', 'hsl2hex', #'hsl2lch',
    'hsv2hsl', 'hsv2hwb', 'hsv2rgb', 'hsv2lab', 'hsv2hex', #'hsv2lch',
    'hwb2hsl', 'hwb2hsv', 'hwb2rgb', 'hwb2lab', 'hwb2hex', #'hwb2lch',
    'rgb2hsl', 'rgb2hsv', 'rgb2hwb', 'rgb2lab', 'rgb2hex', #'rgb2lch',
    'lab2hsl', 'lab2hsv', 'lab2hwb', 'lab2rgb', 'lab2hex', #'lab2lch',
#   'lch2hsl', 'lch2hsv', 'lch2hwb', 'lch2rgb', 'lch2lab', 'lch2hex',
    'hex2hsl', 'hex2hsv', 'hex2hwb', 'hex2rgb', 'hex2lab', #'hex2lch',

    'bytes2norm',
    'norm2bytes',
    'hex2bytes',
    'bytes2hex',

    'rgb2hue',
    'hxx_in_degpct',

    ]


#def hsl2lch(hsl):
#def hwb2lch(hwb):
#def lab2lch(lab):
    


# Range conversions between [0,1] and [0,255]

@colpal_function
def bytes2norm(vec):
    """
    Converts list of unsigned bytes to floats in the range [0,1].
    """
    return [c/255.0 for c in vec]


@colpal_function
def norm2bytes(vec):
    """
    Converts list of floats in the range [0,1] to ints in the range [0,255]. 
    """
    return [ round(x * 255) for x in vec ]



# Conversions between hex strings and bytes [0,255]

@colpal_function
def bytes2hex(vec):
    """
    Converts a list of ints in the range [0,255] to hex strings in #RRGGBB  
    format. 
    """
    return '#' + ''.join('%02X' % i for i in vec)


@hexpal_function
def hex2bytes(hexstr):
    """
    Converts a hex string in #RRGGBB format to a list of ints in [0,255].
    """
    if hexstr[0] == '#':
        hexstr = hexstr[1:]
    return [int(hexstr[:2], 16), int(hexstr[2:4], 16), int(hexstr[4:6], 16)]



# Hex codes
def hex2rgb(hexstr):
    """Converts a hex string in #RRGGBB format to a RGB vec3 in [0,1]. """
    return bytes2norm( hex2bytes( hexstr ))



def hex2hsl(hexcode):
    """Converts hex strings in #RRGGBB format to HSL.  """
    return rgb2hsl( hex2rgb( hexcode ))



def hex2hsv(hexcode):
    """Converts hex strings in #RRGGBB format to HSV.  """
    return rgb2hsv( hex2rgb( hexcode ))



def hex2hwb(hexcode):
    """Converts hex strings in #RRGGBB format to HWB.  """
    return rgb2hwb( hex2rgb( hexcode ))



def hex2lab(hexcode):
    """Converts hex strings in #RRGGBB format to L*a*b*.  """
    return rgb2lab( hex2rgb( hexcode))

#def hex2lch(hexcode):



def hsl2hex(hsl):
    """
    Converts an HSL vec3 to a hex string in #RRGGBB format.
    """
    return bytes2hex(norm2bytes(hsl2rgb( hsl )))



def hsv2hex(hsv):
    """
    Converts an HSV vec3 to a hex string in #RRGGBB format.
    """
    return bytes2hex( norm2bytes( hsv2rgb( hsv )))



def hwb2hex(hwb):
    """
    Converts an HWB vec3 to a hex string in #RRGGBB format.
    """
    return bytes2hex( norm2bytes( hwb2rgb( hwb )))



def lab2hex(lab):
    """
    Converts an HSL vec3 to a hex string in #RRGGBB format.
    """
    return bytes2hex( norm2bytes( lab2rgb( lab )))



def rgb2hex(rgb):
    """
    Converts an HSL vec3 to a hex string in #RRGGBB format.
    """
    return bytes2hex( norm2bytes( rgb ))



### Color mode conversion functions

@colpal_function
def hsv2hsl(hsv):
    """Interprets the given vec3 as HSV and converts it to HSL"""
    h = hsv[0]
    l = hsv[2] * (1-hsv[1]/2)
    s = 0 if l==0 or l==1 else (hsv[2]-l)/min(l, 1-l)
    return [h, s, l]
 
    
          
@colpal_function
def hsl2hsv(hsl):    
    """Interprets the given vec3 as HSL and converts it to HSV"""
    h = hsl[0]
    v = hsl[2] + hsl[1] * min(hsl[2], 1-hsl[2])
    s = 0 if v == 0 else 2 * (1 -hsl[2]/v)
    return [h, s, v]



@colpal_function
def rgb2hsv(rgb):
    """Interprets the given vec3 as RGB and converts it to HSV"""
    return list(colorsys.rgb_to_hsv(*rgb))



@colpal_function
def rgb2hsl(rgb):
    """Interprets the given vec3 as RGB and converts it to HSL"""
    h, l, s = colorsys.rgb_to_hls(*rgb)
    return [h, s, l]



@colpal_function
def hsv2rgb(hsv): 
    """Interprets the given vec3 as HSV and converts it to RGB"""
    return list(colorsys.hsv_to_rgb(*hsv))



@colpal_function
def hsl2rgb(hsl):
    """Interprets the given vec3 as HSL and converts it to RGB"""
    return list(colorsys.hls_to_rgb(hsl[0], hsl[2], hsl[1]))



@colpal_function
def rgb2xyz(rgb):
    """Interprets the given vec3 as RGB and converts it to XYZ"""
    # This is a necessary initial step for RGB->L*a*b*
    # http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    # https://www.easyrgb.com/en/math.php
    tmp = list(rgb)
    for i in range(3):  
        if tmp[i] > 0.04045: 
            tmp[i] = ((tmp[i] + 0.055) / 1.055) ** 2.4
        else: 
            tmp[i] /= 12.92
        
        tmp[i] *= 100
    (r, g, b) = tmp
    # Observer = 2 degrees, Illuminant = D65
    return [
        (r * 0.4124564) + (g * 0.3575761) + (b * 0.1804375),
        (r * 0.2126729) + (g * 0.7151522) + (b * 0.0721750),
        (r * 0.0193339) + (g * 0.1191920) + (b * 0.9503041)]



@colpal_function
def xyz2rgb(xyz):
    """Interprets the given vec3 as XYZ and converts it to RGB"""
    # This is a necessary final step for L*a*b*->RGB
    # http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_RGB.html
    # https://www.easyrgb.com/en/math.php
    (x, y, z) = xyz
    x /= 100.0
    y /= 100.0
    z /= 100.0
    rgb = [ 
        (x * 3.2404542)  + (y * -1.5371385) + (z * -0.4985314),
        (x * -0.9692660) + (y * 1.8760108)  + (z * 0.0415560),
        (x * 0.0556434)  + (y * -0.2040259) + (z * 1.0572252)]
    for i in range(3):
        if rgb[i] > 0.0031308:
            rgb[i] = 1.055 * (rgb[i]**(1.0/2.4)) - 0.055
        else:
            rgb[i] *= 12.92
        rgb[i] = min(max(0.0, rgb[i]), 1.0)
    return rgb



@colpal_function
def xyz2lab(xyz):
    """Interprets the given vec3 as XYZ and converts it to L*a*b*"""
    # https://www.easyrgb.com/en/math.php
    tmp = list(xyz)
    tmp[0] /= 95.047 
    tmp[1] /= 100.000
    tmp[2] /= 108.883
    for i in range(3):
        if tmp[i] > 0.008856:
            tmp[i] **= (1.0/3.0)
        else:
            tmp[i] = (7.787 * tmp[i]) + (16.0/116.0)
    return [
        (116 * tmp[1]) - 16, 
        500 * (tmp[0] - tmp[1]), 
        200 * (tmp[1] - tmp[2])] 
        
    
@colpal_function
def lab2xyz(lab):
    """Interprets the given vec3 as L*a*b* and converts it to XYZ"""
    # https://www.easyrgb.com/en/math.php
    y = (lab[0] + 16) / 116.0
    x = lab[1] / 500.0 + y
    z = y - lab[2] / 200.0
    xyz = [x, y, z]
    for i in range(3):
        cube = xyz[i] ** 3
        if cube > 0.008856: 
            xyz[i] = cube
        else: 
            xyz[i] = (xyz[i] - 16/116.0) / 7.787
    xyz[0] = xyz[0] * 95.047
    xyz[1] = xyz[1] * 100.000
    xyz[2] = xyz[2] * 108.883
    return xyz



def rgb2lab(rgb):
    """Interprets the given vec3 as RGB and converts it to L*a*b*."""
    return xyz2lab(rgb2xyz(rgb))



def lab2rgb(lab):
    """Interprets the given vec3 as L*a*b* and converts it to RGB"""
    return xyz2rgb(lab2xyz(lab))



def lab2hsv(lab):
    """Interprets the given vec3 as L*a*b* and converts it to HSV"""
    return rgb2hsv(lab2rgb(lab))



def lab2hsl(lab):
    """Interprets the given vec3 as L*a*b* and converts it to HSL"""
    return rgb2hsl(lab2rgb(lab))



def hsv2lab(hsv):
    """Interprets the given vec3 as HSV and converts it to L*a*b*"""
    return rgb2lab(hsv2rgb(hsv))



def hsl2lab(hsl):
    """Interprets the given vec3 as HSL and converts it to L*a*b*"""
    return rgb2lab(hsl2rgb(hsl))



@colpal_function
def hsv2hwb(hsv):
    """Interprets the given vec3 as HSV and converts to HWB"""
    return [ hsv[0], (1-hsv[1])*hsv[2], 1-hsv[2] ]



@colpal_function
def hwb2hsv(hwb):
    """Interprets the given vec3 as HWB and converts to HSV"""
    (h,w,b) = hwb
    scalar = w + b
    if scalar > 1:
        w /= scalar
        b /= scalar
    sat = 0.0 if b == 1 else 1-w/(1-b)
    return [ h, sat, 1-b ]



def rgb2hwb(rgb):
    """Interprets the given vec3 as RGB and converts to HWB"""
    return hsv2hwb(rgb2hsv(rgb))



def hwb2rgb(hwb):
    """Interprets the given vec3 as HWB and converts to RGB"""
    return hsv2rgb(hwb2hsv(hwb))



def hsl2hwb(hsl):
    """Interprets the given vec3 as HSL and converts to HWB"""
    return hsv2hwb(hsl2hsv(hsl))



def hwb2hsl(hwb):
    """Interprets the given vec3 as HWB and converts to HSL"""
    return hsv2hsl(hwb2hsv(hwb))



def lab2hwb(lab):
    """Interprets the given vec3 as L*a*b* and converts to HWB"""
    return hsv2hwb(lab2hsv(lab))



def hwb2lab(hwb):
    """Interprets the given vec3 as HWB and converts it to L*a*b* """
    return hsv2lab(hwb2hsv(hwb))




# Miscellaneous conversions. 

@colpal_function
def hxx_in_degpct(hxx):
    (h, c1, c2) = hxx
    err = 10e-8 # TODO take a closer look at this.
    return [ round(360*h+err), round(100*c1+err,3), round(100*c2+err,3) ]


@colpal_function
def rgb2hue(rgb):
    (r,g,b) = rgb
    mn = min(rgb)
    mx = max(rgb)
    delta = mx-mn
    if delta == 0: return 0.0
    multiplier = 1.0/(6*delta)

    if r == mx:
        hue = (g-b) * multiplier
    elif g == mx:
        hue = (1.0/3) + (b-r) * multiplier
    elif b == mx:
        hue = (2.0/3) + (r-g) * multiplier

    if hue < 0: 
        return hue + 1
    elif hue > 1: 
        return hue - 1

    return hue
